from dataclasses import dataclass, field, asdict, fields
from typing import Optional, Dict, Any, List
from omlmd.model_metadata import ModelMetadata
from oras.provider import Registry
from omlmd.provider import OMLMDRegistry
import os
import urllib.request

def write_content_to_file(filename: str, content_fn):
    try:
        with open(filename, 'x') as f:
            content = content_fn()
            f.write(content)
    except FileExistsError:
        raise RuntimeError(f"File '{filename}' already exists. Aborting TODO: demonstrator.")


def download_file(uri):
    file_name = os.path.basename(uri)
    urllib.request.urlretrieve(uri, file_name)
    return file_name


class Helper:
    def __init__(self, registry: Optional[OMLMDRegistry] = None):
        if registry is None:
            self._registry = OMLMDRegistry(insecure=True) # TODO: this is a bit limiting when used from CLI, to be refactored
        else:
            self._registry = registry

    @property
    def registry(self):
        return self._registry

    def push(
        self,
        target: str,
        path: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        author: Optional[str] = None,
        model_format_name: Optional[str] = None,
        model_format_version: Optional[str] = None,
        **kwargs
    ):
        dataclass_fields = {f.name for f in fields(ModelMetadata)} # avoid anything specified in kwargs which would collide
        custom_properties = {k: v for k, v in kwargs.items() if k not in dataclass_fields}
        model_metadata = ModelMetadata(
            name=name,
            description=description,
            author=author,
            customProperties=custom_properties,
            model_format_name=model_format_name,
            model_format_version=model_format_version
        )
        write_content_to_file("model_metadata.omlmd.json", lambda: model_metadata.to_json())
        write_content_to_file("model_metadata.omlmd.yaml", lambda: model_metadata.to_yaml())
        files = [
            f"{path}:application/x-mlmodel",
            "model_metadata.omlmd.json:application/x-config",
            "model_metadata.omlmd.yaml:application/x-config",
        ]
        try:
            # print(target, files, model_metadata.to_annotations_dict())
            return self._registry.push(
                target=target,
                files=files,
                manifest_annotations=model_metadata.to_annotations_dict(),
                manifest_config="model_metadata.omlmd.json:application/x-config"
            )
        finally:
            os.remove("model_metadata.omlmd.json")
            os.remove("model_metadata.omlmd.yaml")


    def pull(
        self,
        target: str,
        outdir: str,
        media_types: Optional[List[str]] = None
    ):
        self._registry.download_layers(target, outdir, media_types)


    def get_config(
        self,
        target: str
    ) -> str:
        return f'{{"reference":"{target}", "config": {self._registry.get_config(target)} }}' # this assumes OCI Manifest.Config later is JSON (per std spec)


    def crawl(
        self,
        targets: List[str]
    ) -> str:
        configs = map(self.get_config, targets)
        joined = "[" + ", ".join(configs) + "]"
        return joined



# Using this to scope CLI targets
import click
from omlmd.helpers import Helper
from omlmd.model_metadata import deserialize_mdfile

@click.group()
def cli():
    pass

@click.command()
@click.argument('target', required=True)
@click.option('-o', '--output', default='.', show_default=True)
@click.option('--media-types', '-m', multiple=True, default=[])
def pull(target, output, media_types):
    """Pulls an OCI Artifact containing ML model and metadata, filtering if necessary."""
    Helper().pull(target, output, media_types)

@click.group()
def get():
    pass

@click.command()
@click.argument('target', required=True)
def config(target):
    """Outputs configuration of the given OCI Artifact for ML model and metadata."""
    click.echo(Helper().get_config(target))

@click.command()
@click.argument('targets', required=True, nargs=-1)
def crawl(targets):
    """Crawls configuration for the given list of OCI Artifact for ML model and metadata."""
    click.echo(Helper().crawl(targets))
    
@click.command()
@click.argument('target', required=True)
@click.argument('path', required=True, type=click.Path())
@click.option('-m', '--metadata', required=True, type=click.Path())
def push(target, path, metadata):
    """Pushes an OCI Artifact containing ML model and metadata, supplying metadata from file as necessary"""
    import logging
    logging.basicConfig(level=logging.DEBUG)
    md = deserialize_mdfile(metadata)
    click.echo(Helper().push(target, path, **md))

cli.add_command(pull)
cli.add_command(get)
get.add_command(config)
cli.add_command(crawl)
cli.add_command(push)



import os
import sys

import oras.defaults
import oras.oci
import oras.provider
import oras.utils
from oras.decorator import ensure_container
import logging
import tempfile

logger = logging.getLogger(__name__)


class OMLMDRegistry(oras.provider.Registry):

    @ensure_container
    def download_layers(self, package, download_dir, media_types):
        """
        Given a manifest of layers, retrieve a layer based on desired media type
        """
        # If you intend to call this function again, you might cache this response
        # for the package of interest.
        manifest = self.get_manifest(package)

        paths = []

        for layer in manifest.get('layers', []):
            if media_types == None or len(media_types) == 0 or layer['mediaType'] in media_types:
                artifact = layer['annotations']['org.opencontainers.image.title']
                outfile = oras.utils.sanitize_path(download_dir, os.path.join(download_dir, artifact))
                path = self.download_blob(package, layer['digest'], outfile)
                paths.append(path)

        return paths
    
    @ensure_container
    def get_config(self, package) -> str:
        """
        Given a manifest of layers, retrieve a layer based on desired media type
        """
        # If you intend to call this function again, you might cache this response
        # for the package of interest.
        manifest = self.get_manifest(package)
        
        manifest_config = manifest.get('config', {})

        for layer in manifest.get('layers', []):
            if layer["digest"] == manifest_config["digest"]:
                temp_dir = tempfile.mkdtemp()
                try:
                    with tempfile.NamedTemporaryFile(dir=temp_dir, mode='w', delete=False) as temp_file:
                        self.download_blob(package, layer['digest'], temp_file.name)
                    with open(temp_file.name, 'r') as temp_file_read:
                        file_content = temp_file_read.read()
                        return file_content
                finally:
                    if os.path.exists(temp_dir):
                        for root, dirs, files in os.walk(temp_dir, topdown=False):
                            for file in files:
                                os.remove(os.path.join(root, file))
                            for dir in dirs:
                                os.rmdir(os.path.join(root, dir))
                        os.rmdir(temp_dir)
                        # print("Temporary directory and its contents have been removed.")



from dataclasses import dataclass, field, asdict
from typing import Optional, Dict, Any
import json
import yaml

@dataclass
class ModelMetadata:
    name: Optional[str] = None
    description: Optional[str] = None
    author: Optional[str] = None
    customProperties: Optional[Dict[str, Any]] = field(default_factory=dict)
    uri: Optional[str] = None
    model_format_name: Optional[str] = None
    model_format_version: Optional[str] = None

    def to_json(self) -> str:
        return json.dumps(asdict(self), indent=4)
    
    def to_dict(self) -> str:
        as_json = self.to_json()
        return json.loads(as_json)
    
    def to_annotations_dict(self) -> str:
        as_dict = self.to_dict()
        result = {}
        # TODO: likely key in annotation require FQDN
        for k, v in as_dict.items():
            if isinstance(v, str):
                result[k] = v
            elif v is None:
                continue
            else:
                result[f"{k}+json"] = json.dumps(v) # post-fix "+json" for OCI annotation which is a str representing a json
        return result

    @staticmethod
    def from_json(json_str: str) -> 'ModelMetadata':
        data = json.loads(json_str)
        return ModelMetadata(**data)

    def to_yaml(self) -> str:
        return yaml.dump(asdict(self), default_flow_style=False)

    @staticmethod
    def from_yaml(yaml_str: str) -> 'ModelMetadata':
        data = yaml.safe_load(yaml_str)
        return ModelMetadata(**data)


def deserialize_mdfile(file):
    with open(file, 'r') as file:
        content = file.read()
        try:
            return json.loads(content)
        except json.JSONDecodeError:
            pass
        
        try:
            return yaml.safe_load(content)
        except yaml.YAMLError:
            pass
        
        raise ValueError(f"The file at {file} is neither a valid JSON nor a valid YAML file.")
